#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/Xutil.h>
#include <math.h>
#include "gfx.h"

struct dttask {
   struct dttask* next;
   void *arg;
};

/* We need to store a couple of things about a button:
   - its coordinates, for click handling
   - its rendering, for repainting it if it gets damaged
   It's best to store all this in a generic 'widget' object
   so that I can later loop over them to see whether they
   accept some event or not. */

enum { WGT_BUTTON };

#define WIDGET_COMMON_FIELDS \
   int type,name; \
   int x,y,w,h; \
   unsigned char *rendered; \
   void (*mouse_press)(int ,int ,int , int); \
   void (*mouse_release)(int, int, int, int);

typedef struct {
  WIDGET_COMMON_FIELDS
} widget_t;

static int nwidgets,swidgets;
static widget_t *widgets;

static struct dttask *dhead;

static Display *Gdpy;
static Window Gwindow;
static int    Gscreen;
static Visual *Gvisual;
static XImage *Gimage;
static XShmSegmentInfo Gshminfo;
static int Gshared;
static Atom wmDelete, wmProtocols;

uint32_t *gfx_pixels;
int       gfx_bpl;
int       gfx_width;
int       gfx_height;

static uint32_t Cfg, Cbg, Caa_fg, Caa_bg;

static gfx_font_func Gbutton_font;

static int  can_use_shared_images() ;
static void create_image(int width,int height) ;
static void create_image_noshm(int width,int height);
static void destroy_image
   (int shared, XShmSegmentInfo *shminfo, XImage* x11image);

static void delayed_shmdt(void *addr);
static void execute_delayed_detachments();

static void paint_screen(int x,int y,int w,int h);
static int dispatch_event(XEvent *event);

static void paint_screen(int x,int y,int w,int h)
{
  if (Gshared) 
     XShmPutImage(Gdpy,Gwindow, DefaultGC(Gdpy,0), Gimage,x,y,x,y,w,h, False);
  else
     XPutImage(Gdpy, Gwindow, DefaultGC(Gdpy,0), Gimage, x, y, x,y, w,h);
}

static void destroy_image
    (int shared, XShmSegmentInfo *shminfo, XImage* x11image)
{
   if (shared)  {
      XShmDetach(Gdpy, shminfo);
      XDestroyImage(x11image);
      delayed_shmdt(shminfo->shmaddr);
   } else {
      XDestroyImage(x11image);
   }
}

void gfx_init(void *disp,int width,int height,const char *title)
{
   int *depths;
   int ndepths;
   int i;
   XVisualInfo xvinfo;
   char *title_str;
   XTextProperty title_prop;

   Gdpy= disp;
   if (!Gdpy)
      Gdpy= XOpenDisplay(NULL);
   if (!Gdpy) 
      exit(fprintf(stderr,"Can't open X11 display. Exiting.\n"));
   Gscreen= XDefaultScreen(Gdpy);

   depths= XListDepths(Gdpy, Gscreen, &ndepths);
   if (!depths) 
     exit(fprintf(stderr, "Can't get list of depths for the "
                          "display. Exiting.\n"));

   for(i=0;i<ndepths;i++)
      if (depths[i]==24) break;
   if (i==ndepths)
      exit(fprintf(stderr, "The display doesn't "
	                   "support 24 bit depth. Exiting.\n"));
   XFree(depths);

   if (XMatchVisualInfo(Gdpy, Gscreen, 24, TrueColor,&xvinfo))
      Gvisual= xvinfo.visual;

   if (!Gvisual &&
       XMatchVisualInfo(Gdpy, Gscreen, 24, DirectColor,&xvinfo))
      Gvisual= xvinfo.visual;

   if (!Gvisual)
      exit(fprintf(stderr,"Can't find a 24 bit "
			  "TrueColor or DirectColor visual. Exiting.\n")); 
	
   if (Gvisual->red_mask!=0xff0000 ||
       Gvisual->green_mask!=0xff00 ||
       Gvisual->blue_mask!=0xff)
      exit(fprintf(stderr,"This program can be used only with RGB\n"
                          "displays with RED_MASK=0xff0000\n"
                          "              GRN_MASK=0x00ff00\n"
                          "              BLU_MASK=0x0000ff\n" 
                          "Other hardware is still possible, but you "
                          "need to write some conversion stuff. Exiting.\n"));

   create_image(width,height);
   Gwindow= XCreateSimpleWindow(Gdpy, DefaultRootWindow(Gdpy), // dpy, parent
                          0, 0, width, height,// xywh
                          0, 0,0);             // border width
                          //24, InputOutput, Gvisual, 0, NULL);
   wmDelete= XInternAtom(Gdpy, "WM_DELETE_WINDOW", True);
   wmProtocols= XInternAtom(Gdpy, "WM_PROTOCOLS", True);

   XSelectInput(Gdpy, Gwindow,
		ButtonPressMask | ButtonReleaseMask |
		StructureNotifyMask | ExposureMask );
   XSetWMProtocols(Gdpy, Gwindow, &wmDelete, 1);
   XSetWindowBackgroundPixmap(Gdpy, Gwindow, None);


   title_str= strdup(title);
   if (!XmbTextListToTextProperty(Gdpy, &title_str, 1, XTextStyle,
                                  &title_prop))
   {
     XSetWMName(Gdpy, Gwindow, &title_prop);
   }

   XMapWindow(Gdpy, Gwindow);

   gfx_width= width;
   gfx_height= height;
}

static void create_image(int width,int height)
{
   struct shmid_ds dummy_shmid;
   size_t pixels_size;

   if (!can_use_shared_images())
      goto fail_nolib;
	
   Gimage= XShmCreateImage(Gdpy,Gvisual,
                        24, ZPixmap,
                        NULL, &Gshminfo,
                        width, height );
   if (!Gimage)
      goto fail_create;
   gfx_bpl= Gimage->bytes_per_line;

   pixels_size= sizeof(uint32_t)*width*height;
	
   Gshminfo.shmid= shmget(IPC_PRIVATE, pixels_size, IPC_CREAT | 0777);
   if (Gshminfo.shmid <0)
      goto fail_shmcreate;
	
   Gshminfo.shmaddr= shmat(Gshminfo.shmid,0,0);

   if (Gshminfo.shmaddr==(char*) -1)
      goto fail_shmat;
	
   gfx_pixels= (unsigned*) Gshminfo.shmaddr;
   Gimage->data= (char*) Gshminfo.shmaddr;
   XShmAttach(Gdpy, &Gshminfo);

   XSync(Gdpy, False);
   /* Doing the XSync after Attach() is important, we give the 
      X Server a chance to attach before we remove the id. Removing
      the id and then telling the server to attach used to work before,
      but no more. Other people had the same problem:
    http://lists.cairographics.org/archives/cairo/2012-October/023650.html
   */
   shmctl(Gshminfo.shmid, IPC_RMID, &dummy_shmid); 

   Gshared= 1;
   return ;
	
fail_shmat: 
   /* do nothing, deletion is already handled */
fail_shmcreate:
   XDestroyImage(Gimage);
fail_create:
fail_nolib:
   create_image_noshm(width,height);
}

static void create_image_noshm(int width,int height)
{
   gfx_pixels= malloc(sizeof(*gfx_pixels)*width*height);
   Gimage= XCreateImage(Gdpy, Gvisual, 24, ZPixmap, 0,
                       (char*) gfx_pixels, width, height, 32,0);
   gfx_bpl= Gimage->bytes_per_line;
   Gshared= 0;
}

static int can_use_shared_images()
{
   static int tried;
   static int can;
   int ignore, major, minor;
   Bool pixmaps;

   if (tried) return can;
   tried= 1;

   if (XQueryExtension(Gdpy, "MIT-SHM", &ignore, &ignore, &ignore) &&
       XShmQueryVersion(Gdpy, &major, &minor, &pixmaps) == True)
      can= 1;

   return can;
}

void gfx_update(int x,int y,int w,int h)
{
   if (w<=0 || h<=0) { x=0; y=0; w= gfx_width; h= gfx_height; }
   XClearArea(Gdpy, Gwindow, x,y, w,h, True);
}

void gfx_close()
{
   destroy_image( Gshared, &Gshminfo, Gimage );
   XCloseDisplay(Gdpy);
   execute_delayed_detachments();
}


static void delayed_shmdt(void *addr)
{
   struct dttask *task;

   task= malloc(sizeof(*task));
   task->arg= addr;
   task->next= dhead;
   dhead= task;
}

static void execute_delayed_detachments()
{
   while(dhead)
   {
      struct dttask *N= dhead->next;
      shmdt(dhead->arg);
      free(dhead);
      dhead= N;
   }
}

static int handle_resize(XEvent *ev)
{
   int W,H;
   W= ev->xconfigure.width;
   H= ev->xconfigure.height;
   if (W==gfx_width && H==gfx_height) return 1;

   int Owidth= gfx_width;
   int Oheight= gfx_height;
   unsigned *Odata= gfx_pixels;
   int Obpl= gfx_bpl;
   int Oshared= Gshared;
   XShmSegmentInfo Oshminfo= Gshminfo;
   XImage *Oimage= Gimage;

   create_image(W,H);
   gfx_width= W;
   gfx_height= H;

   int y;
   int mW= gfx_width;
   int mH= gfx_height;
   if (Owidth<mW) mW= Owidth;
   if (Oheight<mH) mH= Oheight;

   for(y=0;y<mH;y++)
   {
      int *line1p= (int*) ((char*) Odata + y*Obpl);
      int *line2p= (int*) ((char*) gfx_pixels + y*gfx_bpl);
      int x;
      for(x=0;x<mW;x++)
         line2p[x]= line1p[x];
   }

   destroy_image(Oshared, &Oshminfo, Oimage);
   return 0;
}


/* This is probably useless, but just in case somebody 
   redefines this stuff. */
static int translate_mouse_button(int btn)
{
  switch(btn)
  {
   case Button1: return GFXMOD_MOUSE1;
   case Button2: return GFXMOD_MOUSE2;
   case Button3: return GFXMOD_MOUSE3;
   case Button4: return GFXMOD_MOUSE4;
   default:      return GFXMOD_MOUSE5;
  }
}

static int translate_mod_mask(int mask)
{
  int V;
  V= 0;

  if (mask&Button1Mask) V |= (1<<GFXMOD_MOUSE1);
  if (mask&Button2Mask) V |= (1<<GFXMOD_MOUSE2);
  if (mask&Button3Mask) V |= (1<<GFXMOD_MOUSE3);
  if (mask&Button4Mask) V |= (1<<GFXMOD_MOUSE4);
  if (mask&Button5Mask) V |= (1<<GFXMOD_MOUSE5);

  if (mask&ShiftMask)   V |= (1<<GFXMOD_SHIFT);
  if (mask&LockMask)    V |= (1<<GFXMOD_LOCK);
  if (mask&ControlMask) V |= (1<<GFXMOD_CONTROL);

  if (mask&Mod1Mask)   V |= (1<<GFXMOD_MOD1);
  if (mask&Mod2Mask)   V |= (1<<GFXMOD_MOD2);
  if (mask&Mod3Mask)   V |= (1<<GFXMOD_MOD3);
  if (mask&Mod4Mask)   V |= (1<<GFXMOD_MOD4);
  if (mask&Mod5Mask)   V |= (1<<GFXMOD_MOD5);

  return V;
}

void gfx_get_event(int *evcode, int *args)
{
   XEvent event;
again:
   XNextEvent(Gdpy, &event);
   if (dispatch_event(&event)) goto again;

   switch(event.type)
   {
   case ButtonPress:
   case ButtonRelease:
      args[0]= event.xbutton.x;
      args[1]= event.xbutton.y;
      args[2]= translate_mouse_button(event.xbutton.button);
      args[3]= translate_mod_mask(event.xbutton.state);
      if (event.type==ButtonPress) *evcode= GFXEV_MOUSEPRESS;
                              else *evcode= GFXEV_MOUSERELEASE;
      break;
   case ConfigureNotify:
      *evcode= GFXEV_RESIZE;
      args[0]= event.xconfigure.width;
      args[1]= event.xconfigure.height;
      break;
   case ClientMessage:
      if (event.xclient.message_type==wmProtocols
       && event.xclient.format==32)
      {
        if (event.xclient.data.l[0]==wmDelete)
        {
          *evcode= GFXEV_QUITREQUEST;
          break;
        }
      }
      goto again;
   }
}

static int dispatch_event(XEvent *event)
{
   execute_delayed_detachments();
   if (event->xany.window!=Gwindow) return 1;
   switch(event->type)
   {
   case Expose:
        paint_screen(event->xexpose.x, event->xexpose.y,
                     event->xexpose.width, event->xexpose.height);
        return 1;
   case ConfigureNotify:
        return handle_resize(event); 
   // FIXME: add the mouse and keyboard stuff here, for the buttons etc.
   }
   return 0;
}


static int aa_pixels[256];

static double srgb_intensity[256];

static void calculate_intensity()
{
   int i;
   static int intensity_calculated;
   if (intensity_calculated) return ;
   intensity_calculated= 1;
   for(i=0;i<256;i++)
   {
     double C= i/255.0;
     if (C<=0.04045)
      srgb_intensity[i]= C / 12.92;
     else
      srgb_intensity[i]= pow((C+0.055)/(1+0.055) , 2.4);
   }
}

static unsigned int srgb_code(double C)
{
   double K;
   
   if (C<=0.0031308) K= C * 12.92;
   else K= -0.055 + pow(C,1/2.4)*(1+0.055);

   if (K<0.0) K= 0.0;
   else if (K>1.0) K= 1.0;

   return (int) (K*255);
}

static void compute_aa_pixels()
{
   int i;
   if (Caa_fg==Cfg && Caa_bg==Cbg) return ;
   calculate_intensity();

   double Rf = srgb_intensity[(Cfg>>16)&0xff];
   double Gf = srgb_intensity[(Cfg>>8)&0xff];
   double Bf = srgb_intensity[(Cfg>>0)&0xff];

   double Rb = srgb_intensity[(Cbg>>16)&0xff];
   double Gb = srgb_intensity[(Cbg>>8)&0xff];
   double Bb = srgb_intensity[(Cbg>>0)&0xff];

   for(i=0;i<256;i++)
   {
     //double V= srgb_intensity[i];
     double V= i/255.0;
     double Y= 1-V;
     int R=  srgb_code(V*Rf + Y*Rb);
     int G=  srgb_code(V*Gf + Y*Gb);
     int B=  srgb_code(V*Bf + Y*Bb);

     aa_pixels[i]= (R&0xff)<<16 | (G&0xff)<<8 | (B&0xff);
   }

   Caa_fg= Cfg;
   Caa_bg= Cbg;
}


#define XCH(A,B) do { int t= A; A=B; B=t; } while(0)

static void draw_horsline(int width, int x0,int y0, int x1, int y1)
{
   double y,M;
   int dy,x,w,tw,Ws,We;

   if (x0>x1) { XCH(x0,x1); XCH(y0,y1); }
   if (x1<0 || x0>=gfx_width) return ;

   dy= y1-y0;
   M= (double)dy/(x1-x0);

   if (x0<0) { y0+= -x0*M; x0= 0; }
   if (x1>=gfx_width) { y0 -= (x1-gfx_width-1)*M; x1= gfx_width-1; }

   if (dy<0)
   {
      if (y0<0 || y1>=gfx_height) return ;
      if (y1<0) { x1-= y1/M; y1= 0; }
      if (y0>=gfx_height) { x0-= (y0-gfx_height-1)/M; y0= gfx_height-1; }
   } else {
      if (y1<0 || y0>=gfx_height) return ;
      if (y0<0) { x0-= y0/M; y0= 0; }
      if (y1>=gfx_height) { x1-= (y1-gfx_height-1)/M; y1= gfx_height-1; }
   }

   if (x0==x1) return ;

   tw= width / cos(atan(M));
   Ws= -tw/2;
   We= tw+Ws;

   y= y0;
   for(x=x0;x<=x1;x++)
   {
     int yy;
     y0= y+Ws; if (y0<0) y0= 0;
     y1= y+We; if (y1>gfx_height) y1= gfx_height;
     for(yy=y0;yy<y1;yy++)
         gfx_pixels[x+yy*gfx_width]= Cfg;
     y+= M;
   }
}

static void draw_versline(int width, int x0,int y0, int x1, int y1)
{
   double x,M;
   int dx,y,w,tw,Ws,We;

   if (y0>y1) { XCH(y0,y1); XCH(x0,x1); }
   if (y1<0 || y0>=gfx_height) return ;

   dx= x1-x0;
   M= (double)dx/(y1-y0);

   if (y0<0)           { x0+= -y0*M; y0= 0; }
   if (y1>=gfx_height) { x1 -= (y1-gfx_height-1)*M; y1= gfx_height-1; }

   if (dx<0)
   {
      if (x0<0 || x1>=gfx_width) return ;
      if (x1<0) { y1-= x1/M; x1= 0; }
      if (x0>=gfx_width) { y0-= (x0-gfx_width-1)/M; x0= gfx_width-1; }
   } else {
      if (x1<0 || x0>=gfx_width) return ;
      if (x0<0) { y0-= x0/M; x0= 0; }
      if (x1>=gfx_width) { y1-= (x1-gfx_width-1)/M; x1= gfx_width-1; }
   }

   if (y0==y1) return ;

   tw= width / cos(atan(M));
   Ws= -tw/2;
   We= tw+Ws;

   x= x0;
   for(y=y0;y<=y1;y++)
   {
     int xx;
     x0= x+Ws; if (x0<0) x0= 0;
     x1= x+We; if (x1>gfx_width) x1= gfx_width;
     for(xx=x0;xx<x1;xx++) gfx_pixels[xx+y*gfx_width]= Cfg;
     x+= M;
   }
}

static void draw_hor_line(int width, int y, int x0,int x1)
{
   int y0,y1;
   int x;

   y0= y- width/2;
   y1= y+ (width-width/2);
   if (y0<0) y0= 0; 
   if (y1>gfx_height) y1= gfx_height;

   if (y0>=gfx_height || y1<0) return ;
   if (x0>x1) XCH(x0,x1);
   if (x0>=gfx_width || x1<0) return ;

   if (x0<0) x0=0;
   if (x1>gfx_width) x1= gfx_width;

   for(y=y0;y<y1;y++)
     for(x=x0;x<x1;x++)
        gfx_pixels[x+y*gfx_width]= Cfg;
}

static void draw_ver_line(int width, int x, int y0,int y1)
{
   int x0,x1;
   int y;

   x0= x- width/2;
   x1= x+ (width-width/2);
   if (x0<0) x0= 0; 
   if (x1>gfx_width) x1= gfx_width;

   if (x0>=gfx_width || x1<0) return ;
   if (y0>y1) XCH(y0,y1);
   if (y0>=gfx_height || y1<0) return ;

   if (y0<0) y0=0;
   if (y1>gfx_height) y1= gfx_height;

   for(y=y0;y<y1;y++)
     for(x=x0;x<x1;x++)
        gfx_pixels[x+y*gfx_width]= Cfg;
}


void gfx_draw_line(int width, int x0,int y0, int x1, int y1)
{
   int dx,dy;
   dx= abs(x1 - x0);
   dy= abs(y1 - y0);
   if (dx==0)                       draw_ver_line(width, x0, y0, y1); 
   else if (dy==0)                  draw_hor_line(width, y0, x0, x1); 
   else if (abs(x1-x0)>=abs(y1-y0)) draw_horsline(width, x0,y0,x1,y1);
   else                             draw_versline(width, x0,y0,x1,y1);
}



void gfx_fill_rect(int x0,int y0, int x1, int y1)
{
   int x,y;

   if (x0>x1) XCH(x0,x1);
   if (y0>y1) XCH(y0,y1);
   if (y0>=gfx_height || x0>=gfx_width || x1<0 || y1<0) return ;
   if (y0<0) y0= 0;
   if (x0<0) x0= 0;
   if (x1>=gfx_width) x1= gfx_width-1;
   if (y1>=gfx_height) y1= gfx_height-1;

   for(y=y0;y<=y1;y++)
     for(x=x0;x<=x1;x++)
       gfx_pixels[x+y*gfx_width]= Cfg;
}

void gfx_fill_circle(int x0, int y0, int radius)
{
  int x = radius, y = 0;
  int radiusError = 1-x;
 
  if (x0+radius<=0 || y0+radius<=0 ||
      x0-radius>=gfx_width || y0-radius>=gfx_height) return ;

  while(x >= y)
  {
    draw_hor_line(1,  y+y0, -x+x0, x+x0);
    draw_hor_line(1, -y+y0, -x+x0, x+x0);
    draw_ver_line(1,  y+x0, -x+y0, x+y0);
    draw_ver_line(1, -y+x0, -x+y0, x+y0);
 
    y++;
    if(radiusError<0)
	radiusError+=2*y+1;
    else
    {
        x--;
        radiusError+=2*(y-x+1);
    }
  }
}

void gfx_draw_rect(int width, int x0, int y0, int x1, int y1)
{
  draw_hor_line(width, y0, x0, x1);
  draw_hor_line(width, y1, x0, x1);
  draw_ver_line(width, x0, y0, y1);
  draw_ver_line(width, x1, y0, y1);
}

void gfx_fg(uint32_t color) { Cfg= color; } 
void gfx_bg(uint32_t color) { Cbg= color; } 


/* This is from utf-8.c */

/*
UTF-8 Table
8421 8421
110x xxxx   2 bytes    mask= e0  result= c0   min= 80       vmask= 1f
1110 xxxx   3 bytes    mask= f0  result= e0   min= 800      vmask= 0f
1111 0xxx   4 bytes    mask= f8  result= f0   min= 10000    vmask= 07
1111 10xx   5 bytes    mask= fc  result= f8   min= 200000   vmask= 03
1111 110x   6 bytes    mask= fe  result= fc   min= 4000000  vmask= 01

10xx xxxx   cont. byte mask= c0  result= 80   vmask= 3f
*/
int utf8_decode_string
  (const uint8_t *input, size_t ilen, uint32_t *output, size_t *olen)
{
   const uint8_t *end,*start;
   static const uint32_t bit31= (1<<31);
   size_t L;
   int have_errors;
   int i, contb;
   uint8_t C;
   uint32_t outc; uint32_t min;

   end= input+ilen;
   L= 0;
   have_errors= 0;

   if (ilen==0) goto done;

again:
   C= input[0];
   if ( (C&0x80) ==0)    { outc= C; input++; goto out; }
   if ( (C&0xe0) ==0xc0) goto seq2; 
   if ( (C&0xf0) ==0xe0) goto seq3;
   if ( (C&0xf8) ==0xf0) goto seq4;

   // Here, we don't care about the seq5 and seq6 stuff.
   // they are invalid anyway, either overlong or greater
   // then the unicode limit. We just treat them as unrecognized
   // bytes and output them.
invalid_byte:
   outc= input[0] | bit31;
   input++;
   have_errors= 1;
   goto out;

seq2: outc= C & 0x1f;  min=    0x80; contb= 1; goto do_cont;
seq3: outc= C & 0x0f;  min=   0x800; contb= 2; goto do_cont;
seq4: outc= C & 0x07;  min= 0x10000; contb= 3; goto do_cont;

do_cont:
   start= input;
   input++;
   for(i=0;i<contb;i++)
   {
     if ((input==end) || ( (input[0]&0xc0) !=0x80)) goto reject;
     outc= (outc << 6) | (input[0]&0x3f);
     input++;
   }

   if (outc<min || outc>0x10ffff) goto reject; 
   if (outc>=0xd800 && outc<=0xdfff) goto reject;
   if (outc==0xffff || outc==0xfffe) goto reject;

out:
   output[L++]= outc;
   if (input<end) goto again;
done:
   *olen= L;
   return have_errors;
reject:
   input= start;
   goto invalid_byte; 
}

static void draw_chargmap(int x0,int y0, unsigned char *data, int w,int h)
{
   int x,y,sX,eX,sY,eY;
   sX= 0; eX= w;
   sY= 0; eY= h;

   if (x0>=gfx_width || y0>=gfx_height) return ;

   if (x0<0) { sX -= x0; x0= 0; }
   if (y0<0) { sY -= y0; y0= 0; }

   if (x0+w>gfx_width) eX= gfx_width-x0; 
   if (y0+h>gfx_height) eY= gfx_height-y0; 

   if (sX>=eX || sY>= eY) return ;

   compute_aa_pixels();

   for(y=sY;y<eY;y++)
      for(x=sX;x<eX;x++)
         gfx_pixels[ (x+x0) + (y+y0)*gfx_width ] =
           aa_pixels[data[x+y*w]];
}

void gfx_text(gfx_font_func font, const char *str,int x,int y)
{
   size_t L,WL;
   int sW, sH, sA;
   unsigned char *sD;
   uint32_t *wstr;

   L= strlen(str);
   wstr= malloc((L+1)*sizeof(uint32_t));
   utf8_decode_string((uint8_t*) str, L, wstr, &WL);

   (*font)(wstr, WL, &sW, &sH, &sA, &sD);
   free(wstr);
   draw_chargmap(x,y, sD, sW, sH);
   free(sD);
}


void gfx_button_font(gfx_font_func font) { Gbutton_font= font; }

int gfx_button(int x,int y, int w, int h, const char *text, int code)
{
   int nlines;
   unsigned char **lineD;
   const char *S,*P;
   int  *lineW;
   int  lineH,lineA;
   uint32_t *wstr; size_t wL;

   int i,xoffset,yoffset,TH;
   
   if (!Gbutton_font || !text[0]) return ;

   wstr= malloc((strlen(text)+1)*sizeof(uint32_t));
   lineD= NULL;
again:
   nlines= 0;
   S= text;
next_line:
   if (S[0]==0) goto done;
   P= strchr(S,'\n');
   if (!P) P= S+strlen(S);
   if (lineD)
   {
      utf8_decode_string(S, P-S, wstr, &wL);
      (*Gbutton_font)(wstr, wL, lineW+nlines, &lineH, &lineA,
                                &(lineD[nlines]));
   }
   nlines++;
   if (P[0]=='\n') { S=P+1; goto next_line; }

done:
   if (!lineD)
   {
     lineD= malloc(sizeof(char*)*nlines);
     lineW= malloc(sizeof(int)*nlines);
     goto again;
   }
   free(wstr);

   XCH(Cfg,Cbg); 
   gfx_fill_rect(x,y, x+w-1, y+h-1);
   XCH(Cfg,Cbg);
   gfx_draw_rect(2, x,y, x+w-1, y+h-1);
   
   TH= nlines * lineH;
   yoffset= (h-TH)/2;
   for(i=0;i<nlines;i++)
   {
      xoffset= (w-lineW[i])/2;
      draw_chargmap(x+xoffset, y+yoffset, lineD[i], lineW[i], lineH);     
      free(lineD[i]);
      yoffset+= lineH;
   }
   free(lineW); free(lineD);
}
